/*
 * Copyright (c) 2001-2002 The Trustees of Indiana University.  
 *                         All rights reserved.
 * Copyright (c) 1998-2001 University of Notre Dame. 
 *                         All rights reserved.
 * Copyright (c) 1994-1998 The Ohio State University.  
 *                         All rights reserved.
 * 
 * This file is part of the LAM/MPI software package.  For license
 * information, see the LICENSE file in the top level directory of the
 * LAM/MPI source distribution.
 *
 * $Id: mpil_spawn.c,v 1.4.2.2 2002/10/14 18:46:17 brbarret Exp $
 *
 * Program to test MPIL_Spawn (LAM-specific function) with simple
 * arguments.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mpi.h"

#include "lamtest_error.h"


/* Only compile much of this file if we're compiling with LAM/MPI.
   It's highly unlikely that we're going to be using this test suite
   with any other implementation of MPI, but what the hey... this is a
   safe way to do it.  LAM_MPI will be #define'd if we're compiling
   with LAM/MPI, and not defined elsewhere. */

#if LAM_MPI
static void do_parent(char *cmd, int count);
static void do_target(char *argv0, char *argv1, char *argv2,
		      MPI_Comm parent_inter);
static void all_to_all(MPI_Comm intra);
static void free_inter(MPI_Comm inter, int do_free);


/* MPIL_Spawn() splits string arguments into whitespace-deliminted
   arrays of (char*), hence, we have to s/ /_/g the arguments here, as
   compared to the other spawn tests in this directory. */

static char *cmd_argv1 = "this_is_argv_1";
static char *cmd_argv2 = "this_is_argv_2";
static int tag = 201;
#endif /* LAM_MPI */


int 
main(int argc, char *argv[])
{
  int rank, size;
  MPI_Comm parent;
 
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &size);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

#if LAM_MPI
  /* Check to see if we *were* spawned -- because this is a test, we
     can only assume the existence of this one executable.  Hence, we
     both mpirun it and spawn it. */

  parent = MPI_COMM_NULL;
  MPI_Comm_get_parent(&parent);
  if (rank == 0 && parent == MPI_COMM_NULL) {
    printf("*** It is normal to fail this test if the\n\t\"%s\"\n", argv[0]);
    printf("*** executable is not availble on all nodes (e.g., if you do\n");
    printf("*** not have the lamtests suite installed on a networked\n");
    printf("*** filesystem between the nodes)\n");
  }

  if (parent != MPI_COMM_NULL) {
    char *argv1 = 0;
    char *argv2 = 0;
    if (argc > 1)
      argv1 = argv[1];
    if (argc > 2)
      argv2 = argv[2];
    do_target(argv[0], argv1, argv2, parent);
  } else
    do_parent(argv[0], size);
#else
  if (rank == 0) 
    printf("This program checks a LAM/MPI specific function -- MPIL_Spawn -- that existed\n"
	   "before the MPI Forum standardized MPI_Comm_spawn.  Some LAM/MPI users still\n"
	   "actively use MPIL_Spawn, so we still have to test/debug it.  But if you're\n"
	   "using some other implementation of MPI, you can either ignore this message\n"
	   "and/or not run this test program.\n\nMake today an MPI day!\n");
#endif

  /* All done */

  MPI_Finalize();
  return 0;
}


#if LAM_MPI

static void
do_parent(char *cmd, int count)
{
  int *errcode;
  MPI_Comm child_inter;
  MPI_Comm intra;
  char *command;

  command = (char*) malloc(strlen(cmd) + strlen(cmd_argv1) + 
			   strlen(cmd_argv2) + 10);
  sprintf(command, "%s %s %s", cmd, cmd_argv1, cmd_argv2);
  errcode = malloc(sizeof(int) * count);
  MPIL_Spawn(MPI_COMM_WORLD, command, 0, &child_inter);

  /* Now do a simple ping pong to everyone in the child */

  MPI_Intercomm_merge(child_inter, 0, &intra);
  all_to_all(intra);
  MPI_Comm_free(&intra);

  /* Clean up */

  free(command);
  free_inter(child_inter, 1);
  free(errcode);
}


static void
do_target(char *argv0, char *argv1, char *argv2, MPI_Comm parent)
{
  int rank;
  MPI_Comm intra;

  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  /* Check that we got the argv that we expected */

  if (strcmp(argv1, cmd_argv1) != 0)
    lamtest_error(__FILE__, __LINE__, "ERROR: Spawn target rank %d got argv[1]=\"%s\" when expecing \"%s\"\n",
		  rank, argv1, cmd_argv1);
  if (strcmp(argv2, cmd_argv2) != 0)
    lamtest_error(__FILE__, __LINE__, "ERROR: Spawn target rank %d got argv[2]=\"%s\" when expecing \"%s\"\n",
		  rank, argv2, cmd_argv2);

  /* Now merge it down to an intra and do a simple all-to-all to
     everyone in the parent */

  MPI_Intercomm_merge(parent, 0, &intra);
  all_to_all(intra);
  MPI_Comm_free(&intra);

  free_inter(parent, 0);
}


static void
all_to_all(MPI_Comm intra)
{
  int size, rank, i;
  int message;

  MPI_Comm_size(intra, &size);
  MPI_Comm_rank(intra, &rank);

  for (i = 0; i < size; i++) {
    message = -1;
    if (i == rank)
      continue;
    else if (i < rank) {
      MPI_Send(&rank, 1, MPI_INT, i, tag, intra);
      MPI_Recv(&message, 1, MPI_INT, i, tag, intra, MPI_STATUS_IGNORE);
    } else {
      MPI_Recv(&message, 1, MPI_INT, i, tag, intra, MPI_STATUS_IGNORE);
      MPI_Send(&rank, 1, MPI_INT, i, tag, intra);
    }

    if (message != i)
      lamtest_error(__FILE__, __LINE__, "ERROR: rank %d got message %d from comm rank %d; expected %d\n", rank, message, i, i);
  }
}


static void
free_inter(MPI_Comm inter, int do_free)
{
  MPI_Comm intra;
  int size;

  MPI_Comm_size(inter, &size);

  MPI_Intercomm_merge(inter, 0, &intra);
  MPI_Comm_size(inter, &size);
  
  MPI_Comm_free(&intra);
  if (do_free)
    MPI_Comm_free(&inter);
}  

#endif /* LAM_MPI */
