/*
 * Copyright (c) 2001-2002 The Trustees of Indiana University.  
 *                         All rights reserved.
 * Copyright (c) 1998-2001 University of Notre Dame. 
 *                         All rights reserved.
 * Copyright (c) 1994-1998 The Ohio State University.  
 *                         All rights reserved.
 * 
 * This file is part of the LAM/MPI software package.  For license
 * information, see the LICENSE file in the top level directory of the
 * LAM/MPI source distribution.
 *
 * $Id: spawn_multiple.c,v 1.4.2.2 2002/10/14 18:46:17 brbarret Exp $
 *
 * Program to test MPI_Comm_spawn_multiple with simple arguments.
 */

#include <stdio.h>
#include <stdlib.h>
#include "mpi.h"

#include "lamtest_error.h"


static void do_parent(char *cmd, int count);
static void do_target(char *argv0, char *argv1, char *argv2,
		      MPI_Comm parent_inter);
static void all_to_all(MPI_Comm intra);
static void free_inter(MPI_Comm inter, int do_free);


static char *cmd_argv1 = "this is argv 1";
static char *cmd_argv2 = "this is argv 2";
static int tag = 201;


int 
main(int argc, char *argv[])
{
  int rank, size;
  MPI_Comm parent;
 
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &size);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  /* Check to see if we *were* spawned -- because this is a test, we
     can only assume the existence of this one executable.  Hence, we
     both mpirun it and spawn it. */

  parent = MPI_COMM_NULL;
  MPI_Comm_get_parent(&parent);
  if (rank == 0 && parent == MPI_COMM_NULL) {
    printf("*** It is normal to fail this test if the\n\t\"%s\"\n", argv[0]);
    printf("*** executable is not availble on all nodes (e.g., if you do\n");
    printf("*** not have the lamtests suite installed on a networked\n");
    printf("*** filesystem between the nodes)\n");
  }

  if (parent != MPI_COMM_NULL) {
    char *argv1 = 0;
    char *argv2 = 0;
    if (argc > 1)
      argv1 = argv[1];
    if (argc > 2)
      argv2 = argv[2];
    do_target(argv[0], argv1, argv2, parent);
  } else
    do_parent(argv[0], size);

  /* All done */

  MPI_Finalize();
  return 0;
}


static void
do_parent(char *cmd, int count)
{
  MPI_Comm intra;
  MPI_Comm child_inter;
  int *errcode;
  int i;
  int maxprocs[2];
  char *command[2];
  char **spawn_argv[2];
  char *argv0[3];
  char *argv1[3];
  MPI_Info info[2];
  int size;

  maxprocs[0] = count;
  maxprocs[1] = count;
  command[0] = cmd;
  command[1] = cmd;
  spawn_argv[0] = argv0;
  spawn_argv[1] = argv1;
  argv0[0] = cmd_argv1;
  argv0[1] = cmd_argv2;
  argv0[2] = NULL;
  argv1[0] = cmd_argv1;
  argv1[1] = cmd_argv2;
  argv1[2] = NULL;
  info[0] = MPI_INFO_NULL;
  info[1] = MPI_INFO_NULL;

  errcode = (int*) malloc(sizeof(int) * count * 2);
  memset(errcode, -1, count * 2);
  MPI_Comm_spawn_multiple(2, command, spawn_argv, maxprocs, info, 0, 
			  MPI_COMM_WORLD, &child_inter, errcode);

  for (size = i = 0; i < 2; i++)
    size += maxprocs[i];
  for (i = 0; i < size; i++)
    if (errcode[i] != MPI_SUCCESS)
      lamtest_error(__FILE__, __LINE__, "ERROR: MPI_Comm_spawn returned errcode[%d] = %d\n", 
		    i, errcode[i]);

  /* Now do a simple ping pong to everyone in the child */

  MPI_Intercomm_merge(child_inter, 0, &intra);
  all_to_all(intra);
  MPI_Comm_free(&intra);

  /* Clean up */

  free_inter(child_inter, 1);
  free(errcode);
}


static void
do_target(char *argv0, char *argv1, char *argv2, MPI_Comm parent)
{
  int rank;
  MPI_Comm intra;

  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  /* Check that we got the argv that we expected */

  if (strcmp(argv1, cmd_argv1) != 0)
    lamtest_error(__FILE__, __LINE__, "ERROR: Spawn target rank %d got argv[1]=\"%s\" when expecing \"%s\"\n",
		  rank, argv1, cmd_argv1);
  if (strcmp(argv2, cmd_argv2) != 0)
    lamtest_error(__FILE__, __LINE__, "ERROR: Spawn target rank %d got argv[2]=\"%s\" when expecing \"%s\"\n",
		  rank, argv2, cmd_argv2);
  
  /* Now merge it down to an intra and do a simple all-to-all to
     everyone in the parent */

  MPI_Intercomm_merge(parent, 0, &intra);
  all_to_all(intra);
  MPI_Comm_free(&intra);

  free_inter(parent, 0);
}


static void
all_to_all(MPI_Comm intra)
{
  int size, rank, i;
  int message;

  MPI_Comm_size(intra, &size);
  MPI_Comm_rank(intra, &rank);

  for (i = 0; i < size; i++) {
    message = -1;
    if (i == rank)
      continue;
    else if (i < rank) {
      MPI_Send(&rank, 1, MPI_INT, i, tag, intra);
      MPI_Recv(&message, 1, MPI_INT, i, tag, intra, MPI_STATUS_IGNORE);
    } else {
      MPI_Recv(&message, 1, MPI_INT, i, tag, intra, MPI_STATUS_IGNORE);
      MPI_Send(&rank, 1, MPI_INT, i, tag, intra);
    }

    if (message != i)
      lamtest_error(__FILE__, __LINE__, "ERROR: rank %d got message %d from comm rank %d; expected %d\n", rank, message, i, i);
  }
}


static void
free_inter(MPI_Comm inter, int do_free)
{
  MPI_Comm intra;
  int size;

  MPI_Comm_size(inter, &size);

  MPI_Intercomm_merge(inter, 0, &intra);
  MPI_Comm_size(inter, &size);
  
  MPI_Comm_free(&intra);
  if (do_free)
    MPI_Comm_free(&inter);
}  
