/*
 * Copyright (c) 2001-2002 The Trustees of Indiana University.  
 *                         All rights reserved.
 * Copyright (c) 1998-2001 University of Notre Dame. 
 *                         All rights reserved.
 * Copyright (c) 1994-1998 The Ohio State University.  
 *                         All rights reserved.
 * 
 * This file is part of the LAM/MPI software package.  For license
 * information, see the LICENSE file in the top level directory of the
 * LAM/MPI source distribution.
 *
 * $Id: lamtest_error.c,v 1.4.2.1 2002/10/14 18:46:23 brbarret Exp $
 *
 * Simple routines to print and store errors for the LAM test suite
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "mpi.h"

#include "lamtest_error.h"


/*
 * Local variables
 */

char *separator = "<--------------------------------------------------------------------------->";

/*
 * Local routines
 */
static void write_error(const char *mode, const char *file, int line, 
			const char *rank, 
			char *format, ...);


static void
write_error(const char *mode, const char *file, int line, 
	    const char *rank, char *format, ...)
{
  FILE *fp;
  char *filename;
  va_list ap;
  char *env;

  /* Try to output this to the error log file */

  env = getenv("LAM_MPI_TEST_FILEID");
  if (env != NULL) {
    if (rank != NULL) {
      filename = malloc(strlen(env) + strlen(rank) + 32);
      sprintf(filename, "../%s.%s", env, rank);
    } else {
      filename = malloc(strlen(env) + 32);
      sprintf(filename, "../%s", env);
    }
    fp = fopen(filename, "a");

    if (mode != NULL)
      fprintf(fp, "%s:", mode);
    else
      fprintf(fp, "(none):");

    if (rank != NULL)
      fprintf(fp, "%s:", rank);
    else
      fprintf(fp, "(none):");

    fprintf(fp, "%s:%d\n", file, line);

    va_start(ap, format);
    vfprintf(fp, format, ap);
    va_end(ap);
    fprintf(fp, "\n%s\n", separator);
    fclose(fp);

    free(filename);
  }
}


int
lamtest_error(const char *file, int line, char *format, ...)
{
  va_list ap;
  char *rank;
  int ret;

#if LAM_MPI
  /* This is internal to LAM/MPI */
  extern int lam_c2c;
  char *mode = (lam_c2c) ? "c2c" : "lamd";
#else
  char *mode = "none";
#endif

  /* Write the error to the file */

  rank = getenv("LAMRANK");
  if (rank == NULL)
    rank = "(unknown rank)";
  va_start(ap, format);
  write_error(mode, file, line, rank, format, ap);

  /* Print the error to stderr */

  fprintf(stderr, 
	  "ERROR: LAM/MPI MPI_COMM_WORLD rank %s, file %s, line %d (mode %s):\n",
	  rank, file, line, mode);
  ret = vfprintf(stderr, format, ap);
  fprintf(stderr, "\n");
  va_end(ap);

  return ret;
}


void
nonmpi_error(const char *desc, char *format, ...)
{
  va_list ap;

  /* Write the error to the file */

  va_start(ap, format);
  write_error(NULL, desc, 0, NULL, format, ap);
  va_end(ap);
}


void
lamtest_fatal_error(const char *file, int line, char *format, ...)
{
  /* Call the normal routine */

  va_list ap;
  va_start(ap, format);
  lamtest_error(file, line, format, ap);
  va_end(ap);

  /* Now MPI abort */

  MPI_Abort(MPI_COMM_WORLD, 1);
  abort();
}
